// from http://members.chello.at/easyfilter/Bresenham.pdf

function setPixel(x,y)
{
	c2_callFunction("setPixel", [x,y]);
}

function plotLine(x0, y0, x1, y1) {
    let dx = Math.abs(x1 - x0),
        sx = x0 < x1 ? 1 : -1;
    let dy = -Math.abs(y1 - y0),
        sy = y0 < y1 ? 1 : -1;
    let err = dx + dy,
        e2; /* error value e_xy */
    for (;;) {
        /* loop */
        setPixel(x0, y0);
        e2 = 2 * err;
        if (e2 >= dy) {
            /* e_xy+e_x > 0 */
            if (x0 == x1) break;
            err += dy;
            x0 += sx;
        }
        if (e2 <= dx) {
            /* e_xy+e_y < 0 */
            if (y0 == y1) break;
            err += dx;
            y0 += sy;
        }
    }
}

function plotQuadBezierSeg(x0, y0, x1, y1, x2, y2) {
    /* plot a limited quadratic Bezier segment */
    let sx = x2 - x1,
        sy = y2 - y1;
    let xx = x0 - x1,
        yy = y0 - y1,
        xy; /* relative values for checks */
    let dx, dy, err, cur = xx * sy - yy * sx; /* curvature */
    //assert(xx * sx <= 0 && yy * sy <= 0); /* sign of gradient must not change */
    if (sx * sx + sy * sy > xx * xx + yy * yy) {
        /* begin with longer part */
        x2 = x0;
        x0 = sx + x1;
        y2 = y0;
        y0 = sy + y1;
        cur = -cur; /* swap P0 P2 */
    }
    if (cur != 0) {
        /* no straight line */
        xx += sx;
        xx *= sx = x0 < x2 ? 1 : -1; /* x step direction */
        yy += sy;
        yy *= sy = y0 < y2 ? 1 : -1; /* y step direction */
        xy = 2 * xx * yy;
        xx *= xx;
        yy *= yy; /* differences 2nd degree */
        if (cur * sx * sy < 0) {
            /* negated curvature? */
            xx = -xx;
            yy = -yy;
            xy = -xy;
            cur = -cur;
        }
        dx = 4.0 * sy * cur * (x1 - x0) + xx - xy; /* differences 1st degree */
        dy = 4.0 * sx * cur * (y0 - y1) + yy - xy;
        xx += xx;
        yy += yy;
        err = dx + dy + xy; /* error 1st step */
        do {
            setPixel(x0, y0); /* plot curve */
            if (x0 == x2 && y0 == y2) return; /* last pixel -> curve finished */
            y1 = 2 * err < dx; /* save value for test of y step */
            if (2 * err > dy) {
                x0 += sx;
                dx -= xy;
                err += dy += yy;
            } /* x step */
            if (y1) {
                y0 += sy;
                dy -= xy;
                err += dx += xx;
            } /* y step */
        } while (dy < 0 && dx > 0); /* gradient negates -> algorithm fails */
    }
    plotLine(x0, y0, x2, y2); /* plot remaining part to end */
}

function plotQuadBezier(x0, y0, x1, y1, x2, y2) {
    /* plot any quadratic Bezier curve */
    let x = x0 - x1,
        y = y0 - y1;
    let t = x0 - 2 * x1 + x2,
        r;
    if (x * (x2 - x1) > 0) {
        /* horizontal cut at P4? */
        if (y * (y2 - y1) > 0) /* vertical cut at P6 too? */
            if (Math.abs((y0 - 2 * y1 + y2) / t * x) > Math.abs(y)) {
                /* which first? */
                x0 = x2;
                x2 = x + x1;
                y0 = y2;
                y2 = y + y1; /* swap points */
            } /* now horizontal cut at P4 comes first */
        t = (x0 - x1) / t;
        r = (1 - t) * ((1 - t) * y0 + 2.0 * t * y1) + t * t * y2; /* By(t=P4) */
        t = (x0 * x2 - x1 * x1) * t / (x0 - x1); /* gradient dP4/dx=0 */
        x = Math.floor(t + 0.5);
        y = Math.floor(r + 0.5);
        r = (y1 - y0) * (t - x0) / (x1 - x0) + y0; /* intersect P3 | P0 P1 */
        plotQuadBezierSeg(x0, y0, x, Math.floor(r + 0.5), x, y);
        r = (y1 - y2) * (t - x2) / (x1 - x2) + y2; /* intersect P4 | P1 P2 */
        x0 = x1 = x;
        y0 = y;
        y1 = Math.floor(r + 0.5); /* P0 = P4, P1 = P8 */
    }
    if ((y0 - y1) * (y2 - y1) > 0) {
        /* vertical cut at P6? */
        t = y0 - 2 * y1 + y2;
        t = (y0 - y1) / t;
        r = (1 - t) * ((1 - t) * x0 + 2.0 * t * x1) + t * t * x2; /* Bx(t=P6) */
        t = (y0 * y2 - y1 * y1) * t / (y0 - y1); /* gradient dP6/dy=0 */
        x = Math.floor(r + 0.5);
        y = Math.floor(t + 0.5);
        r = (x1 - x0) * (t - y0) / (y1 - y0) + x0; /* intersect P6 | P0 P1 */
        plotQuadBezierSeg(x0, y0, Math.floor(r + 0.5), y, x, y);
        r = (x1 - x2) * (t - y2) / (y1 - y2) + x2; /* intersect P7 | P1 P2 */
        x0 = x;
        x1 = Math.floor(r + 0.5);
        y0 = y1 = y; /* P0 = P6, P1 = P7 */
    }
    plotQuadBezierSeg(x0, y0, x1, y1, x2, y2); /* remaining part */
}